<?php
/*
 Plugin Name: BannerZon
 Plugin URI: http://bannerzon.com
 Description: Quickly and easily create high converting banner ads using Amazon products and categories.
 Version: 1.0.1
 Author: BannerZon
 Author URI: http://bannerzon.com/
 */

if(!class_exists('BannerZon')) {
	class BannerZon {
		/// CONSTANTS

		//// VERSION
		const PLUGIN_SLUG = 'bannerzon';
		const VERSION = '1.0.1';

		//// KEYS
		const AUTO_INSERT_KEY = '_bannerzon_auto_insert';
		const BANNERZON_VIEWS = '_bannerzon_views';
		const BESTSELLER_BASE = '_ba_bs_%s_%s';
		const BROWSE_NODE_CACHE_TRANSIENT_BASE = '_bnh_';
		const ITEM_LOOKUP_BASE = '_ba_il_%s_%s';
		const POST_META_KEY = '_bannerzon_post_meta';
		const SETTINGS_KEY = '_bannerzon_settings';
		const SETTINGS_ERRORS_KEY = '_bannerzon_settings_errors';

		//// MISC
		const ID_BASE = 'bannerzon-banner-%d';

		//// TYPES
		const TYPE_BANNERZON = 'bannerzon';

		//// SLUGS
		const SETTINGS_PAGE_SLUG = 'bannerzon-settings';

		//// CACHE
		const CACHE_PERIOD = 86400; // 24 HOURS

		//// UPDATES AND VALIDATION
		const SERVICE_PRODUCT_ID = '55';
		const SERVICE_VERIFICATION_URL = 'http://bannerzon.com';
		const TRANSIENT_VERIFICATION_CHECK = '_bannerzon_by_key';
		const TRANSIENT_VERIFICATION_CHECK_TIMEOUT = 604800; // 1 Week

		/// DATA STORAGE
		private static $admin_page_hooks = array('media-upload-popup', 'post.php', 'post-new.php', 'edit.php');
		private static $banner_ids = 0;
		private static $default_meta = array();
		private static $default_settings = array();
		private static $first_update_check = true;

		//// URLS
		const URLS__CLICKBANK_SIGNUP = 'https://www.clickbank.com/affiliateAccountSignup.htm';
		const URLS__FAQ = 'http://bannerzon.com/faq/';

		public static function init() {
			self::add_actions();
			self::add_filters();
			self::initialize_defaults();
			self::initialize_thg_amazon_api();
			self::register_shortcodes();

			register_activation_hook(__FILE__, array(__CLASS__, 'do_activation_actions'));
			register_deactivation_hook(__FILE__, array(__CLASS__, 'do_deactivation_actions'));
		}

		private static function add_actions() {
			if(is_admin()) {
				add_action('add_meta_boxes_' . self::TYPE_BANNERZON, array(__CLASS__, 'remove_abberant_meta_boxes'), 10000000);
				add_action('admin_enqueue_scripts', array(__CLASS__, 'enqueue_administrative_resources'));
				add_action('admin_menu', array(__CLASS__, 'add_administrative_interface_items'));
				add_action('admin_print_footer_scripts', array(__CLASS__, 'add_quicktags'));
				add_action('media_upload_bannerzon_ad', array(__CLASS__, 'display_media_upload_bannerzon_ad_iframe'));
			} else {
				add_action('wp_head', array(__CLASS__, 'enqueue_frontend_resources'), 1);
				add_action('wp_head', array(__CLASS__, 'output_boxsizing_script'), 1001);
			}

			add_action('init', array(__CLASS__, 'register_content_types'));
			add_action('manage_' . self::TYPE_BANNERZON . '_posts_custom_column', array(__CLASS__, 'add_columns_output'), 10, 2);
			add_action('parse_request', array(__CLASS__, 'redirect_to_amazon'));
			add_action('save_post', array(__CLASS__, 'save_post_meta'), 10, 2);
			add_action('widgets_init', array(__CLASS__, 'register_widget'));
			add_action('wp_ajax_bannerzon_get_browse_node_options', array(__CLASS__, 'ajax_output_browse_node_options'));
			add_action('wp_ajax_bannerzon_get_item_options', array(__CLASS__, 'ajax_output_item_options'));
			add_action('wp_ajax_bannerzon_views', array(__CLASS__, 'ajax_bannerzon_views'));
			add_action('wp_ajax_nopriv_bannerzon_views', array(__CLASS__, 'ajax_bannerzon_views'));
		}

		private static function add_filters() {
			add_filter('bannerzon_get_settings', array(__CLASS__, 'add_default_category'));
			add_filter('bannerzon_pre_meta_save', array(__CLASS__, 'sanitize_and_validate_meta'));
			add_filter('bannerzon_pre_settings_save', array(__CLASS__, 'sanitize_and_validate_settings'), 10, 2);
			add_filter('generate_rewrite_rules', array(__CLASS__, 'add_rewrite_rules'));
			add_filter('manage_edit-' . self::TYPE_BANNERZON . '_columns', array(__CLASS__, 'add_columns'));
			add_filter('mce_buttons', array(__CLASS__, 'add_tinymce_button'));
			add_filter('mce_external_plugins', array(__CLASS__, 'add_tinymce_plugin'));
			add_filter('plugin_action_links_'.plugin_basename(__FILE__), array(__CLASS__, 'add_settings_link'));
			add_filter('plugins_api', array(__CLASS__, 'check_for_update_info'), 10, 3);
			add_filter('query_vars', array(__CLASS__, 'add_query_vars'));
			add_filter('quicktags_settings', array(__CLASS__, 'add_quicktags_settings'), 10, 2);
			add_filter('site_transient_update_plugins', array(__CLASS__, 'check_for_update'), 20);
			add_filter('the_content', array(__CLASS__, 'auto_insert_banners'));
		}

		private static function initialize_defaults() {
			// Default Settings

			/// Licensing
			self::$default_settings['accept-tos'] = 'no';

			/// Credentials

			/// Links
			self::$default_settings['links-new-window'] = 'yes';
			self::$default_settings['links-nofollow'] = 'yes';
			self::$default_settings['links-cart'] = 'no';
			self::$default_settings['links-tracking'] = 'no';
			self::$default_settings['links-cloaking'] = 'no';
			self::$default_settings['links-cloaking-prefix'] = 'item';

			/// Display and Defaults

			//// General
			self::$default_settings['default-locale'] = 'us';
			self::$default_settings['default-size'] = '300:250';
			self::$default_settings['default-display-price'] = 'no';
			self::$default_settings['default-display-type'] = 'individual';

			//// Design
			self::$default_settings['default-background-color'] = '#ffffff';
			self::$default_settings['default-text-color'] = '#000000';
			self::$default_settings['default-link-color'] = '#0000ee';
			self::$default_settings['default-border-thickness'] = '1';
			self::$default_settings['default-border-style'] = 'solid';
			self::$default_settings['default-border-color'] = '#000000';
			self::$default_settings['default-border-radius'] = '5';

			// Default Meta
			$settings = self::get_settings();

			/// Insertion
			self::$default_meta['auto-insert'] = 'no';
			self::$default_meta['auto-insert-content-types'] = array('post', 'page');
			self::$default_meta['auto-insert-position'] = 'beginning';
			self::$default_meta['auto-insert-align'] = 'center';

			/// Items
			self::$default_meta['locale'] = $settings['default-locale'];
			self::$default_meta['display-type'] = $settings['default-display-type'];
			self::$default_meta['asins'] = array();
			self::$default_meta['category'] = $settings['default-category'];
			self::$default_meta['category-name'] = $settings['default-category-name'];

			/// Design
			self::$default_meta['size'] = $settings['default-size'];

			self::$default_meta['display-price-default'] = 'yes';
			self::$default_meta['display-price'] = $settings['default-display-price'];

			self::$default_meta['background-color-default'] = 'yes';
			self::$default_meta['background-color'] = $settings['default-background-color'];

			self::$default_meta['text-color-default'] = 'yes';
			self::$default_meta['text-color'] = $settings['default-text-color'];

			self::$default_meta['link-color-default'] = 'yes';
			self::$default_meta['link-color'] = $settings['default-link-color'];

			self::$default_meta['border-default'] = 'yes';
			self::$default_meta['border-thickness'] = $settings['default-border-thickness'];
			self::$default_meta['border-style'] = $settings['default-border-style'];
			self::$default_meta['border-color'] = $settings['default-border-color'];

			self::$default_meta['border-radius-default'] = 'yes';
			self::$default_meta['border-radius'] = $settings['default-border-radius'];

			global $wpdb;
			$wpdb->bannerzon_clicks = $wpdb->prefix . 'bannerzon_clicks';
		}


		private static function initialize_thg_amazon_api() {
			$settings = self::get_settings();

			THG_Amazon_API::set_credentials($settings['access-key-id'], $settings['secret-access-key']);
		}

		private static function register_shortcodes() {
			add_shortcode('bannerzon', array(__CLASS__, 'shortcode_output__bannerzon'));
		}

		/// AJAX CALLBACKS

		public static function ajax_bannerzon_views() {
			$data = stripslashes_deep($_REQUEST);

			if(isset($data['bannerzon_views']) && is_array($data['bannerzon_views'])) {
				$results = array();
				foreach($data['bannerzon_views'] as $banner_view_data) {
					$results[$banner_view_data['banner_id']] = self::record_banner_view($banner_view_data['banner_id'], $banner_view_data['count']);
				}

				echo json_encode($results);
				exit;
			}
		}

		public static function ajax_output_browse_node_options() {
			$data = stripslashes_deep($_REQUEST);

			$browse_node_id = $data['browse_node_id'];
			$locale = $data['locale'];
			$hierarchy = array_map('trim', explode(',', $data['hierarchy']));

			$browse_nodes = self::get_browse_nodes($browse_node_id, $locale, $hierarchy);

			foreach ($browse_nodes as $browse_node_id => $browse_node) {
				printf('<option data-ancestral="%s" data-hierarchy="%s" data-children="%d" value="%s">%s</option>', esc_attr($browse_node['ancestral_name']), implode(',', $browse_node['hierarchy']), $browse_node['has_children'], esc_attr($browse_node_id), esc_html($browse_node['name']));
			}

			exit;
		}

		public static function ajax_output_item_options() {
			$data = stripslashes_deep($_REQUEST);

			$locale = $data['locale'];
			$page = isset($data['page']) ? $data['page'] : 1;
			$search_index = $data['search_index'];
			$search_term = $data['search_term'];

			$items_response = self::get_items($locale, $search_index, $search_term, $page);

			ob_start();
			foreach($items_response['items'] as $product) {
				include('views/backend/misc/item-option.php');
			}
			$rows = ob_get_clean();

			$pagination_parts = array();
			if($page > 1) {
				$pagination_parts[] = sprintf('<div class="alignleft"><a class="button button-secondary bannerzon-individual-search-pagination" href="#" data-search-locale="%s" data-search-index="%s" data-search-term="%s" data-page="%d">%s</a></div>',
										$locale,
										$search_index,
										$search_term,
										$page - 1,
										__('Previous'));
			}

			if($page + 1 <= $items_response['pages']) {
				$pagination_parts[] = sprintf('<div class="alignright"><a class="button button-secondary bannerzon-individual-search-pagination" href="#" data-search-locale="%s" data-search-index="%s" data-search-term="%s" data-page="%d">%s</a></div>',
										$locale,
										$search_index,
										$search_term,
										$page + 1,
										__('Next'));
			}

			$pagination = implode('', $pagination_parts) . '<div class="clear"></div>';

			$data = compact('rows', 'pagination');

			echo json_encode($data);

			exit;
		}

		/// CALLBACKS

		public static function add_administrative_interface_items() {
			self::$admin_page_hooks[] = $settings = add_submenu_page(sprintf('edit.php?post_type=%s', self::TYPE_BANNERZON), __('BannerZon Settings'), __('Settings'), 'manage_options', self::SETTINGS_PAGE_SLUG, array(__CLASS__, 'display_settings_page'));

			add_action("load-{$settings}", array(__CLASS__, 'process_settings_save'));
		}

		public static function add_banner_meta_boxes($post) {
			add_meta_box('banner-settings', __('Banner Settings'), array(__CLASS__, 'display_banner_meta_box'), $post->post_type, 'normal');
		}

		public static function add_columns($columns) {
			$settings = self::get_settings();

			$columns['bannerzon-views'] = __('Views');

			if('yes' === $settings['links-tracking']) {
				$columns['bannerzon-clicks'] = __('Clicks');
				$columns['bannerzon-ctr'] = __('CTR');
			}

			return $columns;
		}

		public static function add_columns_output($column, $post_id) {
			switch($column) {
				case 'bannerzon-views':
					echo number_format_i18n((float)get_post_meta($post_id, self::BANNERZON_VIEWS, true), 0);
					break;
				case 'bannerzon-clicks':
					echo number_format_i18n((float)self::get_click_count($post_id), 0);
					break;
				case 'bannerzon-ctr':
					$views = (float)get_post_meta($post_id, self::BANNERZON_VIEWS, true);
					$clicks = self::get_click_count($post_id);

					echo 0 == $views ? 0 : sprintf('%.2f%%', $clicks / $views);
					break;
			}
		}

		public static function add_default_category($settings) {
			if (empty($settings['category']) || empty($settings['category-name'])) {
				$settings['category'] = THG_Amazon_API::get_locale_default_browse_node_id($settings['locale']);
				$settings['category-name'] = THG_Amazon_API::get_locale_default_browse_node_name($settings['locale']);
			}

			return $settings;
		}

		public static function add_meta_boxes($post) {
			add_meta_box('bannerzon-meta-box', __('BannerZon Settings'), array(__CLASS__, 'display_meta_box'), $post->post_type, 'normal');
		}

		public static function add_settings_link($links) {
			$new_links = array();

			$new_links[] = $settings_link = sprintf('<a href="%s">%s</a>', add_query_arg(array('post_type' => self::TYPE_BANNERZON, 'page' => self::SETTINGS_PAGE_SLUG), admin_url('edit.php')), __('Settings'));

			$settings = self::get_settings();
			if('yes' === $settings['links-tracking'] && self::_table_missing()) {
				$new_links[] = $create_table_link = sprintf('<a href="%s">%s</a>', wp_nonce_url(add_query_arg(array('post_type' => self::TYPE_BANNERZON, 'page' => self::SETTINGS_PAGE_SLUG), admin_url('edit.php')), 'bannerzon-create-table'), __('Initialize'));
			}

			return array_merge($new_links, $links);
		}

		public static function add_query_vars($vars) {
			$vars[] = 'bannerzon-asin';
			$vars[] = 'bannerzon-locale';
			$vars[] = 'bannerzon-code';

			return $vars;
		}

		public static function add_rewrite_rules($wp_rewrite) {
			$settings = self::get_settings();

			if('yes' == $settings['links-cloaking']) {
				$prefix = empty($settings['links-cloaking-prefix']) ? 'product' : $settings['links-cloaking-prefix'];

				$new = array(
					"{$prefix}/([^/]+)/([^/]+)(/([^/]+))?/?$" => 'index.php?bannerzon-locale='.$wp_rewrite->preg_index(1).'&bannerzon-asin='.$wp_rewrite->preg_index(2).'&bannerzon-code='.$wp_rewrite->preg_index(4)
				);

				$wp_rewrite->rules = $new + $wp_rewrite->rules;
			}
		}

		public static function add_tinymce_button($buttons) {
			array_push($buttons, '|', 'bannerzon_ad');

			return $buttons;
		}

		public static function add_tinymce_plugin($plugins) {
			$plugins['bannerzon_ad'] = plugins_url('resources/backend/tinymce/editor_plugin.js', __FILE__);

			return $plugins;
		}


		public static function add_quicktags_settings($init, $editor_id) {
			$init['buttons'] .= ',bannerzon_ad';

			return $init;
		}

		public static function add_quicktags() {
			include('views/backend/misc/quicktags.php');
		}

		public static function auto_insert_banners($content) {
			if(is_singular() || is_page()) {

				$post_type = get_post_type();

				$banners = get_posts(array('nopaging' => true, 'post_type' => self::TYPE_BANNERZON, 'meta_key' => self::AUTO_INSERT_KEY, 'meta_value' => 'yes'));
				foreach($banners as $banner) {
					$meta = self::get_meta($banner->ID);

					if(in_array($post_type, $meta['auto-insert-content-types'])) {
						$markup = self::get_banner_markup($banner->ID, $meta['auto-insert-align'], array('context' => 'auto-insert', 'post-id' => get_the_ID(), 'post-type' => $post_type));

						if('beginning' === $meta['auto-insert-position']) {
							$content = "<div class='bannerzon-in-content'>{$markup}</div>" . $content;
						} else if('end' === $meta['auto-insert-position']) {
							$content = $content . "<div class='bannerzon-in-content'>{$markup}</div>";
						}
					}
				}

			}

			return $content;
		}

		public static function check_for_update($checked_data) {
			if (!is_object($checked_data) || !is_array($checked_data->checked) || empty($checked_data->checked)) {
				return $checked_data;
			}

			global $pagenow;
			$should_be_fresh = 'update-core.php' === $pagenow && self::$first_update_check;
			self::$first_update_check = false;

			$version_info = self::get_version_info(self::get_settings('license-key'), $should_be_fresh);
			if (!is_wp_error($version_info) && is_object($version_info) && $version_info->is_valid && version_compare($version_info->new_version, self::VERSION, '>')) {
				$checked_data->response[plugin_basename(__FILE__)] = $version_info;
			}

			return $checked_data;
		}

		public static function check_for_update_info($def, $action, $args) {
			if ($args->slug != self::PLUGIN_SLUG) {
				return $def;
			}

			$update_info = self::get_update_info(self::get_settings('license-key'));

			return $update_info;
		}

		public static function do_activation_actions() {
			self::_create_table();
		}

		public static function do_deactivation_actions() {

		}

		public static function enqueue_administrative_resources($hook) {
			if(!in_array($hook, self::$admin_page_hooks)) { return; }

			wp_enqueue_script('bannerzon-backend', plugins_url('resources/backend/bannerzon.js', __FILE__), array('jquery', 'farbtastic'), self::VERSION);
			wp_enqueue_style('bannerzon-backend', plugins_url('resources/backend/bannerzon.css', __FILE__), array('farbtastic'), self::VERSION);

			wp_localize_script('bannerzon-backend', 'BannerZon_Vars', array(
				'no_browse_node_options_found' => __('Sorry, no categories were found for the action you took.'),
				'bannerzon_ad_shortcode_popup_caption' => __('Insert BannerZon Ad'),
				'bannerzon_ad_shortcode_popup_url' => get_upload_iframe_src('bannerzon_ad')
			));
		}

		public static function enqueue_frontend_resources() {
			wp_enqueue_script('bannerzon-frontend', plugins_url('resources/frontend/bannerzon.js', __FILE__), array('jquery'), self::VERSION);
			wp_enqueue_style('bannerzon-frontend', plugins_url('resources/frontend/bannerzon.css', __FILE__), array(), self::VERSION);

			wp_localize_script('bannerzon-frontend', 'BannerZon_Vars', array(
				'ajaxurl' => admin_url('admin-ajax.php')
			));
		}

		public static function output_boxsizing_script() {
			$boxsizing_url = plugins_url('resources/frontend/boxsizing.htc', __FILE__);

			include('views/frontend/misc/boxsizing.php');
		}

		public static function process_settings_save() {
			$data = stripslashes_deep($_REQUEST);

			if(!empty($data['save-bannerzon-settings-nonce']) && wp_verify_nonce($data['save-bannerzon-settings-nonce'], 'save-bannerzon-settings')) {

				$settings = apply_filters('bannerzon_pre_settings_save', $data['bannerzon'], self::get_settings());
				$settings = self::set_settings($settings);

				if(!is_wp_error(self::get_settings_errors())) {
					add_settings_error('general', 'settings_updated', __('Settings saved.'), 'updated');
					set_transient('settings_errors', get_settings_errors(), 30);
				}

				flush_rewrite_rules();

				wp_redirect(add_query_arg(array('post_type' => self::TYPE_BANNERZON, 'page' => self::SETTINGS_PAGE_SLUG, 'settings-updated' => 'true'), admin_url('edit.php')));
				exit;
			} else if(!empty($data['_wpnonce']) && wp_verify_nonce($data['_wpnonce'], 'bannerzon-create-table')) {
				self::_create_table();

				add_settings_error('general', 'settings_updated', __('Data storage initialized.'), 'updated');
				set_transient('settings_errors', get_settings_errors(), 30);

				wp_redirect(add_query_arg(array('post_type' => self::TYPE_BANNERZON, 'page' => self::SETTINGS_PAGE_SLUG, 'settings-updated' => 'true'), admin_url('edit.php')));
				exit;
			}
		}

		public static function redirect_to_amazon($wp) {
			$data = stripslashes_deep($_REQUEST);
			$vars = $wp->query_vars;

			if($vars['bannerzon-locale'] && $vars['bannerzon-asin']) {
				$asin = $vars['bannerzon-asin'];
				$associate_tag = $vars['bannerzon-code'];
				$locale = $vars['bannerzon-locale'];
				$settings = self::get_settings();

				if('yes' === $settings['links-cart']) {
					$url = self::get_cart_item_url(
						THG_Amazon_API::get_locale_tld($locale),
						$asin,
						$settings['access-key-id'],
						$associate_tag);
				} else {
					$url = sprintf('http://www.amazon.%s/dp/%s', THG_Amazon_API::get_locale_tld($locale), $asin);
					if(!empty($associate_tag)) {
						$url = add_query_arg(array('tag' => $associate_tag), $url);
					}
				}

				if('yes' === $settings['links-tracking']) {
					if(self::_table_missing()) {
						self::_create_table();
					}

					self::record_click($data['banner-id'], $data['product-id'], $data['locale'], $data['attributes']);
				}

				wp_redirect($url);
				exit;
			}
		}

		/**
		 * The client requested we remove non-BannerZon meta boxes from this page.
		 */
		public static function remove_abberant_meta_boxes($post) {
			global $wp_meta_boxes;

			$whitelist = array('submitdiv', 'slugdiv', 'banner-settings');
			foreach($wp_meta_boxes as $screen => $contexts) {
				if(self::TYPE_BANNERZON === $screen) {
					foreach($contexts as $context => $priorities) {
						foreach($priorities as $priority => $ids) {
							foreach($ids as $id => $data) {
								if(!in_array($id, $whitelist)) {
									remove_meta_box($id, $screen, $context);
								}
							}
						}
					}
				}
			}
		}

		public static function register_content_types() {
			$labels = array(
				'name' => __('Amazon Banners'),
				'menu_name' => __('BannerZon'),
				'singular_name' => __('Amazon Banner'),
				'add_new' => __('Add New'),
				'add_new_item' => __('Add New Amazon Banner'),
				'edit_item' => __('Edit Amazon Banner'),
				'new_item' => __('New Amazon Banner'),
				'view_item' => __('View Amazon Banner'),
				'search_items' => __('Search Amazon Banners'),
				'not_found' => __('No Amazon banners found.'),
				'not_found_in_trash' => __('No Amazon banners found in Trash.'),
				'parent_item_colon' => null,
				'all_items' => __('All Amazon Banners'),
			);

			$args = array(
				'labels' => $labels,
				'description' => __('High converting Amazon based banner advertisements.'),
				'publicly_queryable' => false,
				'exclude_from_search' => true,
				'public' => false,
				'rewrite' => false,
				'has_archive' => false,
				'query_var' => false,
				'supports' => array('title'),
				'register_meta_box_cb' => array(__CLASS__, 'add_banner_meta_boxes'),
				'show_ui' => true,
				'menu_position' => null,
				'menu_icon' => null,
				'can_export' => true,
				'show_in_nav_menus' => true,
				'show_in_menu' => true,
				'show_in_admin_bar' => true,
			);

			register_post_type(self::TYPE_BANNERZON, $args);
		}

		public static function register_widget() {
			register_widget('BannerZonWidget');
		}

		public static function sanitize_and_validate_meta($meta) {
			$meta = pd_trim_r($meta);

			// Insertion
			$meta['auto-insert'] = pd_yes_no($meta['auto-insert']);

			if(!is_array($meta['auto-insert-content-types'])) {
				$meta['auto-insert-content-types'] = array();
			}

			$meta['auto-insert-position'] = in_array($meta['auto-insert-position'], array('beginning', 'end')) ? $meta['auto-insert-position'] : 'end';

			// Content
			if(!in_array($meta['locale'], THG_Amazon_API::get_locale_keys())) {
				$meta['locale'] = self::$default_meta['locale'];
			}

			if(!in_array($meta['display-type'], array('individual', 'category'))) {
				$meta['display-type'] = self::$default_meta['display-type'];
			}

			if(!is_array($meta['asins'])) {
				$meta['asins'] = array();
			}

			if(!is_array($meta['asin-links'])) {
				$meta['asin-links'] = array();
			}

			if(!is_array($meta['asin-titles'])) {
				$meta['asin-titles'] = array();
			}

			$meta['asins'] = array_values(array_filter($meta['asins']));
			$meta['asin-links'] = array_values(array_filter($meta['asin-links']));
			$meta['asin-titles'] = array_values(array_filter($meta['asin-titles']));

			if(empty($meta['category'])) {
				$meta['category'] = THG_Amazon_API::get_locale_default_browse_node_id($meta['locale']);
			}

			if (!empty($meta['category'])) {
				$meta['category-name'] = self::get_browse_node_name($meta['category'], $meta['locale']);
			}

			// Design
			$sizes = array_keys(self::get_banner_sizes());
			if(!in_array($meta['size'], $sizes)) {
				$meta['size'] = self::$default_meta['size'];
			}

			$default_keys = array('display-price-default', 'background-color-default', 'text-color-default', 'link-color-default', 'border-default', 'border-radius-default');
			foreach($default_keys as $default_key) {
				$meta[$default_key] = pd_yes_no($meta[$default_key]);
			}



			$meta['display-price'] = in_array($meta['display-price'], array('default', 'yes', 'no')) ? $meta['display-price'] : 'default';

			$color_fields = array('background-color', 'text-color', 'link-color', 'border-color');
			foreach($color_fields as $color_field) {
				$meta[$color_field] = strtolower($meta[$color_field]);

				if(!preg_match('/^#[a-zA-Z0-9]{6}$/', $meta[$color_field])) {
					$meta[$color_field] = self::$default_meta[$color_field];
				}
			}

			$numeric_fields = array('border-thickness', 'border-radius');
			foreach($numeric_fields as $numeric_field) {
				if(!is_numeric($meta[$numeric_field]) || $meta[$numeric_field] < 0) {
					$meta[$numeric_field] = self::$default_meta[$numeric_field];
				}

				$meta[$numeric_field] = intval($meta[$numeric_field]);
			}

			return $meta;
		}

		public static function sanitize_and_validate_settings($settings, $old_settings) {
			$errors = new WP_Error;
			$settings = pd_trim_r($settings);

			$validated = self::validate_license($settings['license-key']);
			if(is_wp_error($validated)) {
				$errors->add('license-key', $validated->get_error_message());
			}

			$settings['accept-tos'] = pd_yes_no($settings['accept-tos']);
			if('yes' !== $settings['accept-tos']) {
				$errors->add('accept-tos', __('You must accept the BannerZon Terms of Service.'));
			}

			if($settings['has-licensed']) {

				/// Credentials
				if(empty($settings['access-key-id']) || empty($settings['secret-access-key'])) {
					if(empty($settings['access-key-id'])) {
						$errors->add('access-key-id', __('Please provide an Access Key ID.'));
					}

					if(empty($settings['secret-access-key'])) {
						$errors->add('secret-access-key', __('Please provide a Secret Access Key.'));
					}
				} else {
					THG_Amazon_API::set_credentials($settings['access-key-id'], $settings['secret-access-key']);
					$browse_node = THG_Amazon_API::browse_node_lookup(THG_Amazon_API::get_locale_default_browse_node_id('us'), null, null, 'us');

					if(is_wp_error($browse_node)) {
						$errors->add('access-key-id', __('Your credentials were incorrect. Please correct them and save the settings again.'));
						$errors->add('secret-access-key', ' ');
					}
				}

				/// Links
				$links_checkboxes = array('links-new-window', 'links-nofollow', 'links-cart', 'links-tracking', 'links-cloaking');
				foreach($links_checkboxes as $links_checkbox) {
					$settings[$links_checkbox] = pd_yes_no($settings[$links_checkbox]);
				}

				//// On the frontend we disable the cloaking checkbox so we need to set the value manually here if this is true
				if('yes' === $settings['links-tracking']) {
					$settings['links-cloaking'] = 'yes';
				}

				$settings['links-cloaking-prefix'] = trim(preg_replace('/[^A-Za-z0-9\-_\/\.]/', '', $settings['links-cloaking-prefix']), '/');
				if(empty($settings['links-cloaking-prefix'])) {
					$settings['links-cloaking-prefix'] = self::$default_settings['links-cloaking-prefix'];
				}

				/// Defaults

				//// Content
				if(!in_array($settings['default-locale'], THG_Amazon_API::get_locale_keys())) {
					$settings['default-locale'] = self::$default_settings['default-locale'];
				}

				if(!in_array($settings['default-display-type'], array('individual', 'category'))) {
					$settings['default-display-type'] = self::$default_settings['default-display-type'];
				}

				if(empty($settings['default-category'])) {
					$settings['default-category'] = THG_Amazon_API::get_locale_default_browse_node_id($settings['default-locale']);
				}

				if (!empty($settings['default-category'])) {
					$settings['default-category-name'] = self::get_browse_node_name($settings['default-category'], $settings['default-locale']);
				}

				//// Design
				$sizes = array_keys(self::get_banner_sizes());
				if(!in_array($settings['default-size'], $sizes)) {
					$settings['default-size'] = self::$default_settings['size'];
				}

				$settings['display-price-default'] = pd_yes_no($settings['display-price-default']);

				$color_fields = array('default-background-color', 'default-text-color', 'default-link-color', 'default-border-color');
				foreach($color_fields as $color_field) {
					$settings[$color_field] = strtolower($settings[$color_field]);

					if(!preg_match('/#[a-zA-Z0-9]{6}/', $settings[$color_field])) {
						$settings[$color_field] = self::$default_settings[$color_field];
					}
				}

				$numeric_fields = array('default-border-thickness', 'default-border-radius');
				foreach($numeric_fields as $numeric_field) {
					if(!is_numeric($settings[$numeric_field]) || $settings[$numeric_field] < 0) {
						$settings[$numeric_field] = self::$default_settings[$numeric_field];
					}

					$settings[$numeric_field] = intval($settings[$numeric_field]);
				}
			} else {
				$settings = array_merge($old_settings, $settings);
			}

			self::set_settings_errors($errors);

			return $settings;
		}

		public static function save_post_meta($post_id, $post) {
			$data = stripslashes_deep($_POST);
			if(wp_is_post_autosave($post_id) || wp_is_post_revision($post_id) || !wp_verify_nonce($data['save-bannerzon-meta-nonce'], 'save-bannerzon-meta')) {
				return;
			}

			$meta = apply_filters('bannerzon_pre_meta_save', $data['bannerzon'], $post_id, $post);
			$meta = self::set_meta($post_id, $meta);

			if('yes' === $meta['auto-insert']) {
				update_post_meta($post_id, self::AUTO_INSERT_KEY, 'yes');
			} else {
				delete_post_meta($post_id, self::AUTO_INSERT_KEY);
			}
		}

		/// DISPLAY CALLBACKS

		public static function get_banner_markup($banner_id, $align, $context_attributes = array()) {
			$banner = get_post($banner_id);
			$css_id = self::get_banner_id();
			$meta = self::get_meta($banner_id);
			$size = $meta['size'];
			$item_count = self::get_banner_item_count($size);
			$link_atts = self::get_link_atts();
			$settings = self::get_settings();

			switch($meta['display-type']) {
				case 'individual':
					$asins = $meta['asins'];
					break;
				case 'category':
					$asins = self::get_bestseller_asins($meta['category'], $meta['locale']);
					break;
			}

			if(empty($asins)) {
				return '';
			}

			shuffle($asins);

			$items = array();
			while(!empty($asins) && count($items) < $item_count) {
				$item = self::get_item($meta['locale'], array_shift($asins));

				if(!empty($item)) {
					$item['CustomURL'] = self::get_item_url($item, $banner, $meta, $css_id, $context_attributes);
					$item['ItemAttributes']['CustomTitle'] = strlen($item['ItemAttributes']['Title']) > 50 ? trim(substr($item['ItemAttributes']['Title'], 0, 50)) . '...' : $item['ItemAttributes']['Title'];
					$items[] = $item;
				}
			}

			$display_price = 'yes' === $meta['display-price-default'] ?
								('yes' === $settings['default-display-price']) :
								('yes' === $meta['display-price']);

			ob_start();
			include('views/frontend/banners/banner.php');
			$markup = ob_get_clean();

			$styles = self::get_banner_css($css_id, $meta);

			return $styles . $markup;
		}

		private static function get_banner_css($css_id, $meta) {
			$settings = self::get_settings();

			$background_color = 'yes' === $meta['background-color-default'] ? $settings['default-background-color'] : $meta['background-color'];
			$text_color = 'yes' === $meta['text-color-default'] ? $settings['default-text-color'] : $meta['text-color'];
			$link_color = 'yes' === $meta['link-color-default'] ? $settings['default-link-color'] : $meta['link-color'];

			$border_thickness = 'yes' === $meta['border-default'] ? $settings['default-border-thickness'] : $meta['border-thickness'];
			$border_color = 'yes' === $meta['border-default'] ? $settings['default-border-color'] : $meta['border-color'];
			$border_style = 'yes' === $meta['border-default'] ? $settings['default-border-style'] : $meta['border-style'];

			$border_radius = 'yes' === $meta['border-radius-default'] ? $settings['default-border-radius'] : $meta['border-radius'];

			$item_width = self::get_banner_item_width($meta['size']);

			ob_start();
			include('views/frontend/banners/styles.php');
			return ob_get_clean();
		}

		public static function display_banner_meta_box($post) {
			$meta = self::get_meta($post->ID);
			$settings = self::get_settings();

			include('views/backend/meta-boxes/banner-meta-box.php');
		}

		public static function display_category_selector($category_name, $category_id, $category_id_input_id, $category_id_input_name, $access_key_id, $secret_access_key, $allow_default = false, $hide_help = false) {
			include ('views/backend/misc/category-selector.php');
		}

		public static function display_media_upload_bannerzon_ad() {
			$ads = new WP_Query(array('nopaging' => true, 'post_type' => self::TYPE_BANNERZON, 'orderby' => 'title', 'order' => 'ASC'));

			include('views/backend/popup/ad-insert.php');
		}

		public static function display_media_upload_bannerzon_ad_iframe() {
			wp_enqueue_style('media');

			return wp_iframe(array(__CLASS__, 'display_media_upload_bannerzon_ad'));
		}

		public static function display_meta_box($post) {
			$meta = self::get_meta($post->ID);

			include('views/backend/meta-boxes/meta-box.php');
		}

		public static function display_settings_page() {
			$errors = self::get_settings_errors();
			$settings = self::get_settings();

			add_meta_box('bannerzon-settings-licensing', __('Licensing'), array(__CLASS__, 'display_settings_meta_box_licensing'), 'bannerzon-settings', 'normal');

			$has_licensed = false;
			$validated = self::validate_license($settings['license-key']);

			if(!is_wp_error($validated) && 'yes' === $settings['accept-tos']) {
				$has_licensed = true;

				add_meta_box('bannerzon-settings-credentials', __('Amazon Credentials'), array(__CLASS__, 'display_settings_meta_box_credentials'), 'bannerzon-settings', 'normal');
				add_meta_box('bannerzon-settings-links', __('Link Settings'), array(__CLASS__, 'display_settings_meta_box_links'), 'bannerzon-settings', 'normal');
				add_meta_box('bannerzon-settings-defaults', __('Defaults'), array(__CLASS__, 'display_settings_meta_box_defaults'), 'bannerzon-settings', 'normal');
			} else {
				if(is_wp_error($validated) && 'yes' !== $settings['accept-tos']) {
					add_settings_error('license-key', 'license-key-invalid', __('Please enter a valid license key and accept the BannerZon Terms of Service to access the rest of the settings.'));
				} else if(is_wp_error($validated)) {
					add_settings_error('license-key', 'license-key-invalid', __('Please enter a valid license key to access the rest of the settings.'));
				} else {
					add_settings_error('license-key', 'license-key-invalid', __('Please accept the BannerZon Terms of Service to access the rest of the settings.'));
				}
			}

			include('views/backend/settings/settings.php');
		}

		//// SETTINGS SECTIONS

		public static function display_settings_meta_box_licensing($data) {
			extract($data);

			include('views/backend/settings/sections/01-licensing.php');
		}

		public static function display_settings_meta_box_credentials($data) {
			extract($data);

			include('views/backend/settings/sections/02-credentials.php');
		}

		public static function display_settings_meta_box_links($data) {
			extract($data);

			$has_permalinks = '' !== get_option('permalink_structure');

			include('views/backend/settings/sections/03-links.php');
		}

		public static function display_settings_meta_box_defaults($data) {
			extract($data);

			include('views/backend/settings/sections/04-defaults.php');
		}

		/// SHORTCODE CALLBACKS

		public static function shortcode_output__bannerzon($atts, $content = null) {
			$atts = shortcode_atts(array('align' => 'none', 'banner_id' => 0), $atts);

			extract($atts);

			if(!empty($banner_id) && self::TYPE_BANNERZON === get_post_type($banner_id)) {
				$align = in_array($align, array('center', 'left', 'right', 'none')) ? $align : 'none';

				$markup = self::get_banner_markup($banner_id, $align, array('context' => 'shortcode', 'post-id' => get_the_ID(), 'post-type' => get_post_type()));
			}

			return "<div class='bannerzon-in-content'>{$markup}</div>";
		}

		/// POST META

		private static function get_meta($post_id) {
			if(empty($post_id)) {
				global $post;
				$post_id = $post->ID;
			}

			$meta = wp_cache_get(self::POST_META_KEY, $post_id);

			if(false === $meta) {
				$meta = wp_parse_args((array)get_post_meta($post_id, self::POST_META_KEY, true), self::$default_meta);
				wp_cache_set(self::POST_META_KEY, $meta, $post_id, time() + self::CACHE_PERIOD);
			}

			return $meta;
		}

		private static function set_meta($post_id, $meta) {
			if(empty($post_id)) {
				global $post;
				$post_id = $post->ID;
			}

			$meta = wp_parse_args($meta, self::$default_meta);
			update_post_meta($post_id, self::POST_META_KEY, $meta);
			wp_cache_set(self::POST_META_KEY, $meta, $post_id, time() + self::CACHE_PERIOD);

			return $meta;
		}

		/// SETTINGS

		private static function get_settings($setting_key = null) {
			$settings = wp_cache_get(self::SETTINGS_KEY);

			if(!is_array($settings)) {
				$settings = apply_filters('bannerzon_get_settings', wp_parse_args(get_option(self::SETTINGS_KEY, self::$default_settings), self::$default_settings));
				wp_cache_set(self::SETTINGS_KEY, $settings, null, time() + self::CACHE_PERIOD);
			}

			return empty($setting_key) ? $settings : (isset($settings[$setting_key]) ? $settings[$setting_key] : false);
		}

		private static function set_settings($settings) {
			if(is_array($settings)) {
				$settings = wp_parse_args($settings, self::$default_settings);
				update_option(self::SETTINGS_KEY, $settings);
				wp_cache_set(self::SETTINGS_KEY, $settings, null, time() + self::CACHE_PERIOD);
			}
		}

		private static function get_settings_errors() {
			$errors = wp_cache_get(self::SETTINGS_ERRORS_KEY);

			if (false === $errors) {
				$errors = get_option(self::SETTINGS_ERRORS_KEY, false);

				if (is_wp_error($errors)) {
					wp_cache_set(self::SETTINGS_ERRORS_KEY, $settings, null, time() + self::CACHE_PERIOD);
				}
			}

			return $errors;
		}

		private static function set_settings_errors($errors) {
			$errors = (is_wp_error($errors) && $errors->get_error_code()) ? $errors : false;

			if (false === $errors) {
				delete_option(self::SETTINGS_ERRORS_KEY);
				wp_cache_delete(self::SETTINGS_ERRORS_KEY);
			} else {
				update_option(self::SETTINGS_ERRORS_KEY, $errors);
				wp_cache_set(self::SETTINGS_ERRORS_KEY, $errors, null, time() + self::CACHE_PERIOD);
			}
		}

		/// AMAZON

		public static function get_bestseller_asins($browse_node_id, $locale) {
			$bestseller_asins_key = sprintf(self::BESTSELLER_BASE, $locale, $browse_node_id);
			$bestseller_asins = get_transient($bestseller_asins_key);

			if (false === $bestseller_asins) {
				$bestseller_items = array();

				$top_sellers_response = THG_Amazon_API::browse_node_lookup($browse_node_id, 'TopSellers', null, $locale);
				if (!is_wp_error($top_sellers_response) && is_array($top_sellers_response['BrowseNodes']['BrowseNode']['TopSellers']['TopSeller'])) {
					$bestseller_asins = array();
					foreach ($top_sellers_response['BrowseNodes']['BrowseNode']['TopSellers']['TopSeller'] as $top_seller) {
						$bestseller_asins[] = $top_seller['ASIN'];
					}

					set_transient($bestseller_asins_key, $bestseller_asins, null, self::CACHE_PERIOD);
				}
			}

			return $bestseller_asins;
		}

		private static function get_browse_nodes($browse_node_id, $locale, $hierarchy = array(-1)) {
			$key = self::BROWSE_NODE_CACHE_TRANSIENT_BASE . $locale . md5($browse_node_id . json_encode($hierarchy));
			$returned = get_transient($key);

			if(!is_array($returned)) {
				$returned = array();

				$queried_browse_node_id = intval($browse_node_id);

				if (0 == $queried_browse_node_id) {
					array_unshift($hierarchy, $queried_browse_node_id);
					$browse_nodes = THG_Amazon_API::get_locale_browse_node_indexes($locale);
					foreach ($browse_nodes as $browse_node_id => $browse_node_name) {
						$returned[$browse_node_id] = array('name' => $browse_node_name, 'ancestral_name' => $browse_node_name, 'has_children' => true, 'hierarchy' => $hierarchy);
					}
				} else {
					$browse_node_info = THG_Amazon_API::browse_node_lookup($browse_node_id, 'BrowseNodeInfo', '', $locale);
					if(!is_wp_error($browse_node_info)) {
						$browse_node_structure = $browse_node_info['BrowseNodes']['BrowseNode'];
						array_unshift($hierarchy, $browse_node_structure['BrowseNodeId']);

						if (is_array($browse_node_structure['Children']['BrowseNode'])) {
							if(isset($browse_node_structure['Children']['BrowseNode']['BrowseNodeId'])) {
								$browse_node_structure['Children']['BrowseNode'] = array($browse_node_structure['Children']['BrowseNode']);
							}

							foreach($browse_node_structure['Children']['BrowseNode'] as $child) {
								if(isset($child['IsCategoryRoot']) && 1 == $child['IsCategoryRoot']) {
									$returned = self::get_browse_nodes($child['BrowseNodeId'], $locale, array_slice($hierarchy, 1));
									set_transient($key, $returned, 23 * 60 * 60);
									return $returned;
								}
							}

							foreach($browse_node_structure['Children']['BrowseNode'] as $child) {
								$returned[$child['BrowseNodeId']] = array(
									'ancestral_name' => self::get_browse_node_ancestral_name($browse_node_structure, array($child['Name'])),
									'has_children' => true,
									'name' => $child['Name'],
									'hierarchy' => $hierarchy,
								);
							}
						}
					} // !is_wp_error
				}

				set_transient($key, $returned, 23 * 60 * 60);
			}

			return $returned;
		}

		private static function get_browse_node_name($browse_node_id, $locale) {
			if ('default' == $browse_node_id) {
				$browse_node_name = __('Site Default');
			} else {
				$browse_node_info = THG_Amazon_API::browse_node_lookup($browse_node_id, '', '', $locale);
				$browse_node_structure = !is_wp_error($browse_node_info) ? $browse_node_info['BrowseNodes']['BrowseNode'] : false;

				if (is_array($browse_node_structure)) {
					$browse_node_name = self::get_browse_node_ancestral_name($browse_node_structure);
				}
			}

			return $browse_node_name;
		}

		private static function get_browse_node_ancestral_name($browse_node_structure, $name_parts = array()) {
			$name_parts = is_array($name_parts) ? $name_parts : array();

			do {
				if(!isset($browse_node_structure['IsCategoryRoot']) || 1 != $browse_node_structure['IsCategoryRoot']) {
					$name_parts[] = $browse_node_structure['Name'];
				}

				$browse_node_structure = $browse_node_structure['Ancestors']['BrowseNode'];
			} while(is_array($browse_node_structure));

			return implode(' > ', array_reverse($name_parts));
		}

		private static function get_items($locale, $search_index, $search_terms, $page = 1) {
			$items_response = THG_Amazon_API::item_search($search_terms, $search_index, intval($page), null, $locale);

			$items = array();
			$pages = 1;
			if(!is_wp_error($items_response)) {
				$items = $items_response['Items']['Item'];
				if(isset($items['ASIN'])) {
					$items = array($items);
				}

				$pages = $items_response['Items']['TotalPages'];
				if($pages > 10) {
					$pages = 10;
				} else if($pages < 1) {
					$pages = 1;
				}
			}

			return compact('items', 'pages');
		}

		private static function get_item($locale, $asin) {
			$item_key = sprintf(self::ITEM_LOOKUP_BASE, $locale, $asin);
			$item = get_transient($item_key);

			if(false === $item) {
				$item_response = THG_Amazon_API::item_lookup($asin, 'ASIN', null, $locale);

				if(!is_wp_error($item_response) && is_array($item_response['Items']['Item'])) {
					$item = $item_response['Items']['Item'][0];
					$item['FetchedTimestamp'] = current_time('timestamp');
					set_transient($item_key, $item, null, self::CACHE_PERIOD);
				}
			}

			return $item;
		}

		private static function get_item_url($item, $banner, $meta, $css_id, $context_attributes) {
			$settings = self::get_settings();

			if('yes' === $settings['links-cloaking']) {
				$prefix = empty($settings['links-cloaking-prefix']) ? 'product' : $settings['links-cloaking-prefix'];

				$url = sprintf(home_url('/%s/%s/%s/'), $prefix, $meta['locale'], $item['ASIN']);

				if(!empty($settings['associates'][$meta['locale']])) {
					$url .= sprintf('%s/', $settings['associates'][$meta['locale']]);
				}

				$vars = array(
					'banner-id' => $banner->ID,
					'product-id' => $item['ASIN'],
					'locale' => $meta['locale'],
					'attributes' => $context_attributes
				);

				$url = add_query_arg($vars, $url);
			} else {
				if('yes' === $settings['links-cart']) {
					$url = self::get_cart_item_url(
						THG_Amazon_API::get_locale_tld($meta['locale']),
						$item['ASIN'],
						$settings['access-key-id'],
						$settings['associates'][$meta['locale']]);
				} else {
					$tag = empty($settings['associates'][$meta['locale']]) ? false : $settings['associates'][$meta['locale']];

					$base = urldecode($item['DetailPageURL']);

					if(false === $tag) {
						$url = remove_query_arg('tag', $base);
					} else {
						$url = add_query_arg(array('tag' => $tag), $base);
					}
				}
			}

			return $url;
		}

		private static function get_cart_item_url($tld, $asin, $access_key_id, $associate_tag) {
			return sprintf('http://www.amazon.%s/gp/aws/cart/add.html?ASIN.1=%s&Quantity.1=1&AWSAccessKeyId=%s&AssociateTag=%s',
				$tld,
				$asin,
				$access_key_id,
				$associate_tag
			);
		}

		/// UTILITY

		private static function get_banner_id() {
			self::$banner_ids++;

			return sprintf(self::ID_BASE, self::$banner_ids);
		}

		private static function get_banner_item_count($size) {
			switch($size) {
				case '120-240':
					$count = 1;
					break;
				case '200-200':
				case '250-250':
				case '468-60':
					$count = 2;
					break;
				case '120-600':
				case '160-600':
				case '300-250':
				case '336-280':
				case '728-90':
					$count = 3;
					break;
				case 'none':
					$count = 4;
					break;
			}

			return $count;
		}

		private static function get_banner_item_width($size) {
			switch($size) {
				case '300-250':
				case '336-280':
				case '250-250':
				case '120-600':
				case '160-600':
				case '200-200':
				case '120-240':
					$width = '100%';
					break;
				case '468-60':
					$width = '50%';
					break;
				case '728-90':
					$width = '33.3%';
					break;
			}

			return $width;
		}

		private static function get_banner_sizes() {
			return array(
				'300-250' => __('300 x 250'),
				'336-280' => __('336 x 280'),
				'250-250' => __('250 x 250'),
				'120-600' => __('120 x 600'),
				'160-600' => __('160 x 600'),
				'468-60'  => __('468 x 60'),
				'728-90'  => __('728 x 90'),
				'200-200' => __('200 x 200'),
				'120-240' => __('120 x 240'),
			);
		}

		private static function get_link_atts() {
			$settings = self::get_settings();

			$atts = array();
			if('yes' === $settings['links-new-window']) {
				$atts[] = 'target="_blank"';
			}

			if('yes' === $settings['links-nofollow']) {
				$atts[] = 'rel="nofollow"';
			}

			return implode(' ', $atts);
		}

		private static function record_banner_view($banner_id, $increment) {
			$views = 0;

			if(!empty($banner_id) && self::TYPE_BANNERZON === get_post_type($banner_id)) {
				$views = intval(get_post_meta($banner_id, self::BANNERZON_VIEWS, true)) + intval($increment);
				update_post_meta($banner_id, self::BANNERZON_VIEWS, $views);
			}

			return $views;
		}

		/// LICENSING

		private static function validate_license($license_key) {
			if(empty($license_key)) {
				$result = new WP_Error('invalid_license_key', __('You must provide a non-empty license key.'));
			} else {
				if (defined('BANNERZON_DEV') && BANNERZON_DEV) {
					return defined('BANNERZON_DEV_VALID_KEY') ? (BANNERZON_DEV_VALID_KEY ? true : new WP_Error('license-key', __('Your license key is invalid.'))) : true;
				}

				$license_key = trim($license_key);
				$transient_key = '_bz_l_' . md5(self::TRANSIENT_VERIFICATION_CHECK . "_{$license_key}");
				$response = get_transient($transient_key);

				if (empty($response) || (is_object($response) && (!isset($response->is_valid) || false === $response->is_valid))) {
					$request_url = add_query_arg(array('clickbank-receipt-id' => urlencode($license_key), 'clickbank-product-id' => self::SERVICE_PRODUCT_ID), self::SERVICE_VERIFICATION_URL);
					$response = wp_remote_get($request_url);
				}

				if (is_wp_error($response)) {
					$result = true;
				} else {
					$object = @json_decode(wp_remote_retrieve_body($response));

					if (is_object($object)) {
						set_transient($transient_key, $response, self::TRANSIENT_VERIFICATION_CHECK_TIMEOUT);
					}

					if (!is_object($object) || $object->is_valid) {
						$result = true;
					} else {
						$result = new WP_Error('invalid_license_key', sprintf(__('Your license key is invalid. Please enter a valid license key.'), add_query_arg(array('post_type' => self::TYPE_BANNERZON, 'page' => self::SETTINGS_PAGE_SLUG), admin_url('edit.php'))));
					}
				}
			}

			return $result;
		}

		/// UPDATES

		private static function get_version_info($license_key, $fresh = false) {
			$license_key = trim($license_key);

			$transient_key = '_bz_i_'.md5(self::TRANSIENT_VERIFICATION_CHECK . $license_key);

			$response = $fresh ? false : get_transient($transient_key);
			if (empty($response) || (is_object($response) && (!isset($response->is_valid) || false == $response->is_valid))) {
				$request_url = add_query_arg(array('check-for-update' => 'yes', 'clickbank-receipt-id' => urlencode($license_key), 'clickbank-product-id' => self::SERVICE_PRODUCT_ID), self::SERVICE_VERIFICATION_URL);
				$response = wp_remote_get($request_url);
				set_transient($transient_key, $response, 60*60*24);
			}

			if (!is_wp_error($response)) {
				$object = @json_decode(wp_remote_retrieve_body($response));

				if (is_object($object) && $object->is_valid) {
					$result = $object;
					$result->slug = self::PLUGIN_SLUG;
				} else {
					$result = new WP_Error;
				}
			} else {
				$result = $response;
			}

			return $result;
		}

		private static function get_update_info($license_key) {
			$update_info = self::get_version_info($license_key);

			if (is_wp_error($update_info)) {
				$data = $update_info;
			} else {
				$data = new stdClass;

				$latest_package = get_object_vars($update_info);

				$data->slug = self::PLUGIN_SLUG;
				$data->version = $latest_package['version'];
				$data->last_updated = $latest_package['date'];
				$data->download_link = $latest_package['package'];
				$data->author = $latest_package['author'];
				$data->external = $latest_package['external'];
				$data->requires = $latest_package['requires'];
				$data->tested = $latest_package['tested'];
				$data->homepage = $latest_package['homepage'];
				$data->downloaded = $latest_package['downloaded'];
				$data->sections = array_map('wpautop', get_object_vars($latest_package['sections']));
			}

			return $data;
		}

		/// CLICKS

		private static function get_click_count($banner_id) {
			global $wpdb;

			return $wpdb->get_var($wpdb->prepare("SELECT COUNT(*) FROM {$wpdb->bannerzon_clicks} WHERE banner_id = %d", $banner_id));
		}

		private static function record_click($banner_id, $product_id, $locale, $attributes) {
			global $wpdb;

			$attributes = maybe_serialize((array)$attributes);

			$wpdb->insert($wpdb->bannerzon_clicks, compact('banner_id', 'product_id', 'locale', 'attributes'), array('%d', '%s', '%s', '%s'));

			return $wpdb->insert_id;
		}

		/// DB

		private static function _create_table() {
			global $charset, $wpdb;

			$queries = array();

			$queries[] = "CREATE TABLE $wpdb->bannerzon_clicks (
						 click_id bigint(20) unsigned NOT NULL auto_increment,
						 banner_id bigint(20) unsigned NOT NULL,
						 product_id varchar(127) NOT NULL default '',
						 locale varchar(10) NOT NULL default '',
						 attributes text NOT NULL default '',
						 PRIMARY KEY  (click_id),
						 KEY banner_id (banner_id)
						) $charset_collate;";


			if(!function_exists('dbDelta')) {
				require_once(path_join(ABSPATH, 'wp-admin/includes/upgrade.php'));
			}

			dbDelta($queries);
		}

		private static function _table_missing() {
			global $wpdb;

			return $wpdb->bannerzon_clicks !== $wpdb->get_var("SHOW TABLES LIKE '{$wpdb->bannerzon_clicks}'");
		}

		/// TEMPLATE TAGS

		public static function get_best_price($item) {
			$price = __('N/A');

			if(isset($item['OfferSummary'])
		    && isset($item['OfferSummary']['LowestNewPrice'])
		    && isset($item['OfferSummary']['LowestNewPrice']['FormattedPrice'])) {

				$price = $item['OfferSummary']['LowestNewPrice']['FormattedPrice'];

			} else if(isset($item['ItemAttributes']['ListPrice'])
				   && isset($item['ItemAttributes']['ListPrice']['FormattedPrice'])) {

				$price = $item['ItemAttributes']['ListPrice']['FormattedPrice'];

			} else if(isset($item['OfferSummary'])
				   && isset($item['OfferSummary']['LowestUsedPrice'])
				   && isset($item['OfferSummary']['LowestUsedPrice']['FormattedPrice'])) {

				$price = $item['OfferSummary']['LowestUsedPrice']['FormattedPrice'];

			} else if(isset($item['Offers'])
				   && isset($item['Offers']['Offer'])
				   && isset($item['Offers']['Offer']['OfferListing'])
				   && isset($item['Offers']['Offer']['OfferListing']['Price'])
				   && isset($item['Offers']['Offer']['OfferListing']['Price']['FormattedPrice'])) {

				$price = $item['Offers']['Offer']['OfferListing']['Price']['FormattedPrice'];

			}

			return $price;
		}

	}

	require_once('lib/amazon-api.php');
	require_once('lib/template-tags.php');
	require_once('lib/utility.php');
	require_once('widget/widget.php');

	BannerZon::init();
}
